@@ -129,8 +129,8 @@ ioctl_tree_last_sibling(ioctl_tree * node)
 /**
  * ioctl_tree_insert:
  *
- * If an equal node already exists, return that node and do not insert "node".
- * Otherwise, insert it and return NULL.
+ * Insert the node and return the (new) root of the tree. If an equal node
+ * already exist, the node will not be inserted.
  */
 ioctl_tree *
 ioctl_tree_insert(ioctl_tree * tree, ioctl_tree * node)
@@ -143,7 +143,7 @@ ioctl_tree_insert(ioctl_tree * tree, ioctl_tree * node)
     if (tree == NULL) {
 	node->last_added = ioctl_node_list_new();
 	ioctl_node_list_append(node->last_added, node);
-	return NULL;
+	return node;
     }
 
     /* trying to insert into itself? */
@@ -153,7 +153,8 @@ ioctl_tree_insert(ioctl_tree * tree, ioctl_tree * node)
     if (existing) {
 	DBG(DBG_IOCTL_TREE, "ioctl_tree_insert: node of type %s ptr %p already exists\n", node->type->name, node);
 	ioctl_node_list_append(tree->last_added, existing);
-	return existing;
+	ioctl_tree_free(node);
+	return tree;
     }
 
     node->parent = node->type->insertion_parent(tree, node);
@@ -178,7 +179,7 @@ ioctl_tree_insert(ioctl_tree * tree, ioctl_tree * node)
     }
 
     ioctl_node_list_append(tree->last_added, node);
-    return NULL;
+    return tree;
 }
 
 ioctl_tree *
@@ -349,8 +350,12 @@ ioctl_tree_execute(ioctl_tree * tree, ioctl_tree * last, IOCTL_REQUEST_TYPE id,
 
     DBG(DBG_IOCTL_TREE, "ioctl_tree_execute ioctl %X\n", (unsigned) id);
 
-    /* check if it's a hardware independent stateless ioctl */
     t = ioctl_type_get_by_id(id);
+    /* Ignore dummy entries that only exist for size resolution. */
+    if (t && t->execute == NULL)
+	t = NULL;
+
+    /* check if it's a hardware independent stateless ioctl */
     if (t != NULL && t->insertion_parent == NULL) {
 	DBG(DBG_IOCTL_TREE, "  ioctl_tree_execute: stateless\n");
 	if (t->execute(NULL, id, arg, &r))
@@ -461,7 +466,8 @@ write_hex(FILE * file, const char *buf, size_t len)
  *
  ***********************************/
 
-#define NSIZE(node) ((node->type && node->type->real_size >= 0) ? node->type->real_size : _IOC_SIZE(node->id))
+#define TSIZE(type, id) ((type && type->real_size >= 0) ? type->real_size : _IOC_SIZE(id))
+#define NSIZE(node) TSIZE(node->type, node->id)
 
 static inline int
 id_matches_type(IOCTL_REQUEST_TYPE id, const ioctl_type *type)
@@ -837,7 +843,7 @@ ioctl_insertion_parent_stateless(ioctl_tree * tree, ioctl_tree * node)
 
 /* ioctl which does not need to consider its data argument and has no state */
 #define I_NOSTATE(name, execute_result) \
-    {name, -1, 0, #name, NULL, NULL, NULL, NULL, NULL, ioctl_execute_ ## execute_result, NULL}
+    {name, 0, 0, #name, NULL, NULL, NULL, NULL, NULL, ioctl_execute_ ## execute_result, NULL}
 
 /* input structs with fixed length with explicit size */
 #define I_NAMED_SIZED_SIMPLE_STRUCT_IN(name, namestr, size, nr_range, insertion_parent_fn) \
@@ -867,20 +873,25 @@ ioctl_insertion_parent_stateless(ioctl_tree * tree, ioctl_tree * node)
 
 /* data with custom handlers; necessary for structs with pointers to nested
  * structs, or keeping stateful handlers */
-#define I_CUSTOM(name, nr_range, fn_prefix)                     \
-    {name, -1, nr_range, #name,                                 \
+#define I_CUSTOM(name, size, nr_range, fn_prefix)               \
+    {name, size, nr_range, #name,                  \
      fn_prefix ## _init_from_bin, fn_prefix ## _init_from_text, \
      fn_prefix ## _free_data,                                   \
      fn_prefix ## _write, fn_prefix ## _equal,                  \
      fn_prefix ## _execute, fn_prefix ## _insertion_parent}
 
+#define I_DUMMY(name, size, nr_range)               \
+    {name, size, nr_range, #name,                  \
+     NULL, NULL, NULL, NULL, NULL, NULL, NULL}     \
+
 ioctl_type ioctl_db[] = {
     I_SIMPLE_STRUCT_IN(USBDEVFS_CONNECTINFO, 0, ioctl_insertion_parent_stateless),
 
     /* we assume that every SUBMITURB is followed by a REAPURB and that
      * ouput EPs don't change the buffer, so we ignore USBDEVFS_SUBMITURB */
-    I_CUSTOM(USBDEVFS_REAPURB, 0, usbdevfs_reapurb),
-    I_CUSTOM(USBDEVFS_REAPURBNDELAY, 0, usbdevfs_reapurb),
+    I_DUMMY(USBDEVFS_SUBMITURB, -1, 0),
+    I_CUSTOM(USBDEVFS_REAPURB, -1, 0, usbdevfs_reapurb),
+    I_CUSTOM(USBDEVFS_REAPURBNDELAY, -1, 0, usbdevfs_reapurb),
 #ifdef USBDEVFS_GET_CAPABILITIES
     I_SIMPLE_STRUCT_IN(USBDEVFS_GET_CAPABILITIES, 0, ioctl_insertion_parent_stateless),
 #endif
@@ -934,6 +945,15 @@ ioctl_type_get_by_id(IOCTL_REQUEST_TYPE id)
     return NULL;
 }
 
+const int ioctl_data_size_by_id(IOCTL_REQUEST_TYPE id)
+{
+    ioctl_type *cur;
+    for (cur = ioctl_db; cur->name[0] != '\0'; ++cur)
+	if (id_matches_type(id, cur))
+	    return TSIZE(cur, id);
+    return 0;
+}
+
 const ioctl_type *
 ioctl_type_get_by_name(const char *name, IOCTL_REQUEST_TYPE *out_id)
 {
