@@ -422,187 +422,6 @@ netlink_recvmsg(int sockfd, struct msghdr * msg, int flags, ssize_t ret)
 }
 
 
-/********************************
- *
- * ioctl recording
- *
- ********************************/
-
-/* global state for recording ioctls */
-int ioctl_record_fd = -1;
-FILE *ioctl_record_log;
-ioctl_tree *ioctl_record;
-struct sigaction orig_actint;
-
-static void ioctl_record_sigint_handler(int signum);
-
-static void
-ioctl_record_open(int fd)
-{
-    libc_func(fopen, FILE*, const char *, const char*);
-    static dev_t record_rdev = (dev_t) - 1;
-
-    if (fd < 0)
-	return;
-
-    /* lazily initialize record_rdev */
-    if (record_rdev == (dev_t) - 1) {
-	const char *dev = getenv("UMOCKDEV_IOCTL_RECORD_DEV");
-
-	if (dev != NULL)
-	    record_rdev = parse_dev_t(dev, "$UMOCKDEV_IOCTL_RECORD_DEV", 1);
-	else {
-	    /* not recording */
-	    record_rdev = 0;
-	}
-    }
-
-    if (record_rdev == 0)
-	return;
-
-    /* check if the opened device is the one we want to record */
-    if (dev_of_fd(fd) != record_rdev)
-	return;
-
-    /* recording is already in progress? */
-    if (ioctl_record_fd >= 0) {
-	/* libmtp opens the device multiple times, we can't do that */
-	/*
-       fprintf(stderr, "umockdev: recording for this device is already ongoing, stopping recording of previous open()\n");
-       ioctl_record_close();
-       */
-       fprintf(stderr, "umockdev: WARNING: ioctl recording for this device is already ongoing on fd %i, but application opened it a second time on fd %i without closing\n", ioctl_record_fd, fd);
-    }
-
-    ioctl_record_fd = fd;
-
-    /* lazily open the record file */
-    if (ioctl_record_log == NULL) {
-	int r;
-
-	const char *path = getenv("UMOCKDEV_IOCTL_RECORD_FILE");
-	const char *device_path = getenv("UMOCKDEV_IOCTL_RECORD_DEVICE_PATH");
-	struct sigaction act_int;
-
-	if (path == NULL) {
-	    fprintf(stderr, "umockdev: $UMOCKDEV_IOCTL_RECORD_FILE not set\n");
-	    exit(1);
-	}
-	if (device_path == NULL) {
-	    fprintf(stderr, "umockdev: $UMOCKDEV_IOCTL_RECORD_DEVICE_PATH not set\n");
-	    exit(1);
-	}
-	if (getenv("UMOCKDEV_DIR") != NULL) {
-	    fprintf(stderr, "umockdev: $UMOCKDEV_DIR cannot be used while recording\n");
-	    exit(1);
-	}
-	ioctl_record_log = _fopen(path, "a+");
-	if (ioctl_record_log == NULL) {
-	    perror("umockdev: failed to open ioctl record file");
-	    exit(1);
-	}
-	/* We record the device node for later loading without specifying
-	 * the devpath in umockdev_testbed_load_ioctl.
-	 */
-	fseek(ioctl_record_log, 0, SEEK_END);
-	if (ftell(ioctl_record_log) > 0) {
-	    /* We're updating a previous log; don't write the devnode header again,
-	     * but check that we're recording the same device as the previous log.
-	     */
-	    char *existing_device_path;
-	    char c;
-	    DBG(DBG_IOCTL, "ioctl_record_open: Updating existing record for path %s\n", path);
-	    fseek(ioctl_record_log, 0, SEEK_SET);
-
-	    /* Start by skipping any leading comments */
-	    while ((c = fgetc(ioctl_record_log)) == '#')
-		while (fgetc(ioctl_record_log) != '\n')
-		    ;
-	    ungetc(c, ioctl_record_log);
-
-	    if (fscanf(ioctl_record_log, "@DEV %ms\n", &existing_device_path) == 1)
-	    {
-		/* We have an existing "@DEV /dev/something" directive, check it matches */
-		DBG(DBG_IOCTL, "ioctl_record_open: recording %s, existing device spec in record %s\n", device_path, existing_device_path);
-		if (strcmp(device_path, existing_device_path) != 0) {
-		    fprintf(stderr, "umockdev: attempt to record two different devices to the same ioctl recording\n");
-		    exit(1);
-		}
-		free(existing_device_path);
-	    }
-
-	    /* load an already existing log */
-	    fseek(ioctl_record_log, 0, SEEK_SET);
-	    ioctl_record = ioctl_tree_read(ioctl_record_log);
-	} else {
-	    /* New log, add devnode header */
-	    DBG(DBG_IOCTL, "ioctl_record_open: Starting new record %s\n", path);
-	    fprintf(ioctl_record_log, "@DEV %s\n", device_path);
-	}
-
-	/* ensure that we write the file also on Control-C */
-	act_int.sa_handler = ioctl_record_sigint_handler;
-	r = sigemptyset(&act_int.sa_mask);
-	assert(r == 0);
-	act_int.sa_flags = 0;
-	r = sigaction(SIGINT, &act_int, &orig_actint);
-	assert(r == 0);
-
-	DBG(DBG_IOCTL, "ioctl_record_open: starting ioctl recording of fd %i into %s\n", fd, path);
-    } else {
-	DBG(DBG_IOCTL, "ioctl_record_open: ioctl recording is already ongoing, continuing on new fd %i\n", fd);
-    }
-}
-
-static void
-ioctl_record_close(int fd)
-{
-    if (fd < 0 || fd != ioctl_record_fd)
-	return;
-
-    DBG(DBG_IOCTL, "ioctl_record_close: stopping ioctl recording on fd %i\n", fd);
-    ioctl_record_fd = -1;
-
-    /* recorded anything? */
-    if (ioctl_record != NULL) {
-	int r;
-
-	rewind(ioctl_record_log);
-	r = ftruncate(fileno(ioctl_record_log), 0);
-	assert(r == 0);
-	fprintf(ioctl_record_log, "@DEV %s\n", getenv("UMOCKDEV_IOCTL_RECORD_DEVICE_PATH"));
-	ioctl_tree_write(ioctl_record_log, ioctl_record);
-	fflush(ioctl_record_log);
-    }
-}
-
-static void ioctl_record_sigint_handler(int signum)
-{
-    int r;
-
-    DBG(DBG_IOCTL, "ioctl_record_sigint_handler: got signal %i, flushing record\n", signum);
-    ioctl_record_close(ioctl_record_fd);
-    r = sigaction(SIGINT, &orig_actint, NULL);
-    assert(r == 0);
-    raise(signum);
-}
-
-static void
-record_ioctl(IOCTL_REQUEST_TYPE request, void *arg, int result)
-{
-    ioctl_tree *node;
-
-    assert(ioctl_record_log != NULL);
-
-    node = ioctl_tree_new_from_bin(request, arg, result);
-    if (node == NULL)
-	return;
-    ioctl_tree_insert(ioctl_record, node);
-    /* handle initial node */
-    if (ioctl_record == NULL)
-	ioctl_record = node;
-}
-
 /********************************
  *
  * ioctl emulation
@@ -612,93 +431,239 @@ record_ioctl(IOCTL_REQUEST_TYPE request, void *arg, int result)
 static fd_map ioctl_wrapped_fds;
 
 struct ioctl_fd_info {
-    ioctl_tree *tree;
-    ioctl_tree *last;
+    char *dev_path;
+    int ioctl_sock;
+    int is_default;
+    pthread_mutex_t sock_lock;
 };
 
 static void
-ioctl_emulate_open(int fd, const char *dev_path)
+ioctl_emulate_open(int fd, const char *dev_path, int must_exist)
 {
-    libc_func(fopen, FILE*, const char *, const char*);
-    libc_func(fclose, int, FILE *);
-    FILE *f;
-    static char ioctl_path[PATH_MAX];
+    libc_func(socket, int, int, int, int);
+    libc_func(connect, int, int, const struct sockaddr *, socklen_t);
+    int is_default = 0;
+    int sock;
+    int ret;
     struct ioctl_fd_info *fdinfo;
+    struct sockaddr_un addr;
 
     if (strncmp(dev_path, "/dev/", 5) != 0)
 	return;
 
-    fdinfo = malloc(sizeof(struct ioctl_fd_info));
-    fdinfo->tree = NULL;
-    fdinfo->last = NULL;
-    fd_map_add(&ioctl_wrapped_fds, fd, fdinfo);
+    addr.sun_family = AF_UNIX;
+    snprintf(addr.sun_path, sizeof(addr.sun_path), "%s/ioctl/%s", getenv("UMOCKDEV_DIR"), dev_path);
 
-    /* check if we have an ioctl tree for this */
-    snprintf(ioctl_path, sizeof(ioctl_path), "%s/ioctl/%s", getenv("UMOCKDEV_DIR"), dev_path);
-
-    f = _fopen(ioctl_path, "r");
-    if (f == NULL)
-	return;
-
-    fdinfo->tree = ioctl_tree_read(f);
-    _fclose(f);
-    if (fdinfo->tree == NULL) {
-	fprintf(stderr, "ERROR: libumockdev-preload: failed to load ioctl record file for %s: empty or invalid format?",
-		dev_path);
-	exit(1);
+    if (path_exists (addr.sun_path) != 0) {
+	snprintf(addr.sun_path, sizeof(addr.sun_path), "%s/ioctl/_default", getenv("UMOCKDEV_DIR"));
+	is_default = 1;
     }
-    DBG(DBG_IOCTL, "ioctl_emulate_open fd %i (%s): loaded ioctl tree\n", fd, dev_path);
+
+    sock = _socket(AF_UNIX, SOCK_STREAM, 0);
+    if (sock == -1) {
+	if (must_exist) {
+	    fprintf(stderr, "ERROR: libumockdev-preload: Failed to open ioctl socket for %s",
+		    dev_path);
+	    exit(1);
+	} else {
+	    return;
+	}
+    }
+
+    ret = _connect(sock, (const struct sockaddr *) &addr, sizeof(addr));
+    if (ret == -1) {
+	if (must_exist) {
+	    fprintf(stderr, "ERROR: libumockdev-preload: Failed to connect to ioctl socket for %s",
+		    dev_path);
+	    exit(1);
+	} else {
+	    return;
+	}
+    }
+
+    fdinfo = malloc(sizeof(struct ioctl_fd_info));
+    fdinfo->ioctl_sock = sock;
+    fdinfo->dev_path = strdup(dev_path);
+    fdinfo->is_default = is_default;
+    pthread_mutex_init(&fdinfo->sock_lock, NULL);
+
+    fd_map_add(&ioctl_wrapped_fds, fd, fdinfo);
+    DBG(DBG_IOCTL, "ioctl_emulate_open fd %i (%s): connected ioctl sockert\n", fd, dev_path);
 }
 
 static void
 ioctl_emulate_close(int fd)
 {
+    libc_func(close, int, int);
     struct ioctl_fd_info *fdinfo;
 
     if (fd_map_get(&ioctl_wrapped_fds, fd, (const void **)&fdinfo)) {
 	DBG(DBG_IOCTL, "ioctl_emulate_close: closing ioctl socket fd %i\n", fd);
 	fd_map_remove(&ioctl_wrapped_fds, fd);
-	ioctl_tree_free(fdinfo->tree);
+	if (fdinfo->ioctl_sock >= 0)
+	    _close(fdinfo->ioctl_sock);
+	free(fdinfo->dev_path);
+	pthread_mutex_destroy(&fdinfo->sock_lock);
 	free(fdinfo);
     }
 }
 
+#define IOCTL_REQ_IOCTL 1
+#define IOCTL_REQ_RES 2
+#define IOCTL_REQ_READ 7
+#define IOCTL_REQ_WRITE 8
+#define IOCTL_RES_DONE 3
+#define IOCTL_RES_RUN 4
+#define IOCTL_RES_READ_MEM 5
+#define IOCTL_RES_WRITE_MEM 6
+#define IOCTL_RES_ABORT 0xff
+
+/* Marshal everything as unsigned long */
+struct ioctl_request {
+    unsigned long cmd;
+    unsigned long arg1;
+    unsigned long arg2;
+};
+
 static int
-ioctl_emulate(int fd, IOCTL_REQUEST_TYPE request, void *arg)
+remote_emulate(int fd, int cmd, long arg1, long arg2)
 {
-    ioctl_tree *ret;
-    int ioctl_result = -1;
-    int orig_errno;
+    libc_func(send, ssize_t, int, const void *, size_t, int);
+    libc_func(recv, ssize_t, int, const void *, size_t, int);
+    libc_func(ioctl, int, int, IOCTL_REQUEST_TYPE, ...);
+    libc_func(read, ssize_t, int, void *, size_t);
+    libc_func(write, ssize_t, int, void *, size_t);
     struct ioctl_fd_info *fdinfo;
+    struct ioctl_request req;
+    sigset_t sig_set, sig_restore;
+    int res;
+
+    /* Block all signals while we are talking with the remote process. */
+    sigfillset(&sig_set);
+    pthread_sigmask(SIG_SETMASK, &sig_set, &sig_restore);
 
     IOCTL_LOCK;
 
-    if (fd_map_get(&ioctl_wrapped_fds, fd, (const void **)&fdinfo)) {
-	/* we default to erroring and an appropriate error code before
-	 * tree_execute, as handlers might change errno; if they succeed, we
-	 * reset errno */
-	orig_errno = errno;
-	/* evdev ioctls default to ENOENT; FIXME: record that instead of
-	 * hardcoding, and handle in ioctl_tree */
-	if (_IOC_TYPE(request) == 'E')
-	    errno = ENOENT;
-	else
-	    errno = ENOTTY;
+    if (!fd_map_get(&ioctl_wrapped_fds, fd, (const void **)&fdinfo)) {
+	IOCTL_UNLOCK;
 
-	/* check our ioctl tree */
-	ret = ioctl_tree_execute(fdinfo->tree, fdinfo->last, request, arg, &ioctl_result);
-	DBG(DBG_IOCTL, "ioctl_emulate: tree execute ret %p, result %i, errno %i (%m); orig errno: %i\n", ret, ioctl_result, errno, orig_errno);
-	if (ret != NULL)
-	    fdinfo->last = ret;
-	if (ioctl_result != -1 && errno != 0)
-	    errno = orig_errno;
-    } else {
-	ioctl_result = UNHANDLED;
+	pthread_sigmask(SIG_SETMASK, &sig_restore, NULL);
+	return UNHANDLED;
     }
-
     IOCTL_UNLOCK;
 
-    return ioctl_result;
+    /* Only pass on ioctl requests for the default handler. */
+    if (fdinfo->is_default && cmd != IOCTL_REQ_IOCTL) {
+	pthread_sigmask(SIG_SETMASK, &sig_restore, NULL);
+	return UNHANDLED;
+    }
+
+    pthread_mutex_lock (&fdinfo->sock_lock);
+
+    /* We force "unsigned int" here to prevent sign extension to long
+     * which could confuse the receiving side. */
+    req.cmd = cmd;
+    req.arg1 = arg1;
+    req.arg2 = arg2;
+
+    res = _send(fdinfo->ioctl_sock, &req, sizeof(req), 0);
+    if (res < 0)
+	goto con_err;
+
+    while (1) {
+	res = _recv(fdinfo->ioctl_sock, &req, sizeof(req), 0);
+	if (res < 0)
+	    goto con_err;
+	if (res == 0)
+	    goto con_eof;
+
+	switch (req.cmd) {
+	    case IOCTL_RES_DONE:
+		errno = req.arg2;
+
+		pthread_mutex_unlock (&fdinfo->sock_lock);
+		pthread_sigmask(SIG_SETMASK, &sig_restore, NULL);
+		/* Force a context switch so that other threads can take the lock */
+		usleep(0);
+		return req.arg1;
+
+	    case IOCTL_RES_RUN:
+		if (cmd == IOCTL_REQ_IOCTL)
+		    res = _ioctl(fd, arg1, arg2);
+		else if (cmd == IOCTL_REQ_READ)
+		    res = _read(fd, (char*) arg1, arg2);
+		else if (cmd == IOCTL_REQ_WRITE)
+		    res = _write(fd, (char*) arg1, arg2);
+		else
+		    goto con_err;
+
+		req.cmd = IOCTL_REQ_RES;
+		req.arg1 = res;
+		req.arg2 = errno;
+
+		res = _send(fdinfo->ioctl_sock, &req, sizeof(req), 0);
+		if (res < 0)
+		    goto con_err;
+
+		break;
+
+	    case IOCTL_RES_READ_MEM: {
+		size_t done = 0;
+		do {
+		    res = _send(fdinfo->ioctl_sock, (void*) (req.arg1 + done), (size_t) req.arg2 - done, 0);
+		    if (res > 0)
+			done += res;
+		} while (res > 0 && done < req.arg2);
+		if (res < 0 && errno == EFAULT) {
+		    fprintf(stderr, "ERROR: libumockdev-preload: emulation code requested invalid read from %p + %lx\n",
+			    (void*) req.arg1, (unsigned long) req.arg2);
+		}
+		if (res < 0)
+		    goto con_err;
+
+		break;
+	    }
+
+	    case IOCTL_RES_WRITE_MEM: {
+		size_t done = 0;
+		do {
+		    res = _recv(fdinfo->ioctl_sock, (void*) (req.arg1 + done), (size_t) req.arg2 - done, 0);
+		    if (res > 0)
+			done += res;
+		} while (res > 0 && done < req.arg2);
+		if (res < 0 && errno == EFAULT) {
+		    fprintf(stderr, "ERROR: libumockdev-preload: emulation code requested invalid write to %p + %lx\n",
+			    (void*) req.arg1, (unsigned long) req.arg2);
+		}
+
+		if (res < 0)
+		    goto con_err;
+
+		break;
+	    }
+
+	    case IOCTL_RES_ABORT:
+		fprintf(stderr, "ERROR: libumockdev-preload: Server requested abort on device %s, exiting\n",
+			fdinfo->dev_path);
+		abort();
+
+	    default:
+		fprintf(stderr, "ERROR: libumockdev-preload: Error communicating with ioctl socket, unknown command: %ld (res: %d)\n",
+			req.cmd, res);
+		abort();
+	}
+    }
+
+con_eof:
+    fprintf(stderr, "ERROR: libumockdev-preload: Error communicating with ioctl socket, received EOF\n");
+    pthread_sigmask(SIG_SETMASK, &sig_restore, NULL);
+    abort();
+
+con_err:
+    fprintf(stderr, "ERROR: libumockdev-preload: Error communicating with ioctl socket, errno: %d\n",
+	    errno);
+    pthread_sigmask(SIG_SETMASK, &sig_restore, NULL);
+    abort();
 }
 
 /********************************
@@ -932,10 +897,9 @@ script_record_open(int fd)
 static void
 script_record_connect(int sockfd, const struct sockaddr *addr, socklen_t addrlen, int res)
 {
-    const char *path = getenv("UMOCKDEV_DIR");
     size_t i;
 
-    if (addr->sa_family == AF_UNIX && res == 0 && path == NULL) {
+    if (addr->sa_family == AF_UNIX && res == 0) {
 	const char *sock_path = ((struct sockaddr_un *) addr)->sun_path;
 
 	/* find out where we log it to */
@@ -1291,12 +1255,9 @@ int prefix ## open ## suffix (const char *path, int flags, ...)	    \
     } else							    \
 	ret =  _ ## prefix ## open ## suffix(p, flags);		    \
     TRAP_PATH_UNLOCK;						    \
-    if (path != p)						    \
-	ioctl_emulate_open(ret, path);				    \
-    else {							    \
-	ioctl_record_open(ret);					    \
+    ioctl_emulate_open(ret, path, path != p);			    \
+    if (path == p)						    \
 	script_record_open(ret);				    \
-    }								    \
     return ret;							    \
 }
 
@@ -1315,12 +1276,9 @@ int prefix ## open ## suffix (const char *path, int flags)	    \
     DBG(DBG_PATH, "testbed wrapped " #prefix "open" #suffix "(%s) -> %s\n", path, p); \
     ret =  _ ## prefix ## open ## suffix(p, flags);		    \
     TRAP_PATH_UNLOCK;						    \
-    if (path != p)						    \
-	ioctl_emulate_open(ret, path);				    \
-    else {							    \
-	ioctl_record_open(ret);					    \
+    ioctl_emulate_open(ret, path, path != p);			    \
+    if (path == p)						    \
 	script_record_open(ret);				    \
-    }								    \
     return ret;						    	    \
 }
 
@@ -1342,10 +1300,8 @@ FILE* prefix ## fopen ## suffix (const char *path, const char *mode)  \
     TRAP_PATH_UNLOCK;						    \
     if (ret != NULL) {						    \
 	int fd = fileno(ret);					    \
-	if (path != p)						    \
-	    ioctl_emulate_open(fd, path);			    \
-	else {							    \
-	    ioctl_record_open(fd);				    \
+	ioctl_emulate_open(fd, path, path != p);		    \
+	if (path == p) {					    \
 	    script_record_open(fd);				    \
 	}							    \
     }								    \
@@ -1498,6 +1454,11 @@ read(int fd, void *buf, size_t count)
     libc_func(read, ssize_t, int, void *, size_t);
     ssize_t res;
 
+    res = remote_emulate(fd, IOCTL_REQ_READ, (long) buf, (long) count);
+    if (res != UNHANDLED) {
+	DBG(DBG_IOCTL, "ioctl fd %i read of %d bytes: emulated, result %i\n", fd, (int) count, (int) res);
+	return res;
+    }
     res = _read(fd, buf, count);
     script_record_op('r', fd, buf, res);
     return res;
@@ -1509,6 +1470,11 @@ write(int fd, const void *buf, size_t count)
     libc_func(write, ssize_t, int, const void *, size_t);
     ssize_t res;
 
+    res = remote_emulate(fd, IOCTL_REQ_WRITE, (long) buf, (long) count);
+    if (res != UNHANDLED) {
+	DBG(DBG_IOCTL, "ioctl fd %i write of %d bytes: emulated, result %i\n", fd, (int) count, (int) res);
+	return res;
+    }
     res = _write(fd, buf, count);
     script_record_op('w', fd, buf, res);
     return res;
@@ -1646,7 +1612,6 @@ close(int fd)
 
     netlink_close(fd);
     ioctl_emulate_close(fd);
-    ioctl_record_close(fd);
     script_record_close(fd);
 
     return _close(fd);
@@ -1660,7 +1625,6 @@ fclose(FILE * stream)
     if (fd >= 0) {
 	netlink_close(fd);
 	ioctl_emulate_close(fd);
-	ioctl_record_close(fd);
 	script_record_close(fd);
     }
 
@@ -1684,19 +1648,16 @@ ioctl(int d, IOCTL_REQUEST_TYPE request, ...)
     arg = va_arg(ap, void*);
     va_end(ap);
 
-    result = ioctl_emulate(d, request, arg);
+    result = remote_emulate(d, IOCTL_REQ_IOCTL, (unsigned int) request, (long) arg);
     if (result != UNHANDLED) {
 	DBG(DBG_IOCTL, "ioctl fd %i request %X: emulated, result %i\n", d, (unsigned) request, result);
 	return result;
     }
 
-    /* call original ioctl */
+    /* fallback to call original ioctl */
     result = _ioctl(d, request, arg);
     DBG(DBG_IOCTL, "ioctl fd %i request %X: original, result %i\n", d, (unsigned) request, result);
 
-    if (result != -1 && ioctl_record_fd == d)
-	record_ioctl(request, arg, result);
-
     return result;
 }
 
